<?php

/*
  This class implements stemming algorithm by Dr. Porter (coded by Anu Joseph-
  CDLR, Strathclyde University, London, UK). Create an instance of the class and call the methods.
  For instance:

  $string =strtolower($string) //string to be stemmed
  $stm = new stemming();
  $stm->reset();
  for($i=0; $i<strlen($string); $i++)
  $stm->add(substr($string, $i, 1));
  $stm->stem(0);
  $stem = $stm->toString(); //stem of the string
*/

class stemming {
  var  $buffer;
  var  $stemIdx; // index of stem
  var  $j;
  var  $k;

  function stemming() {
    $this->buffer = array();
    $this->stemIdx = 0;
  }

  /* cons(i) is true <=> b[i] is a consonant. */
  function cons($i){
    switch ($this->buffer[$i]){
    case "a":
    case "e":
    case "i":
    case "o":
    case "u":
      return false;
    case "y":
      return ($i==0) ? true : !($this->cons($i-1));
    default:
      return true;
    }
  }

  /* m() measures the number of consonant sequences between 0 and j. if c is
     a consonant sequence and v a vowel sequence, and <..> indicates arbitrary
     presence,

     <c><v>       gives 0
     <c>vc<v>     gives 1
     <c>vcvc<v>   gives 2
     <c>vcvcvc<v> gives 3
     .... */

  function m(){
    $n = 0;
    $i = 0;
    while(true){
      if ($i > $this->j)
        return $n;
      if (! $this->cons($i))
        break;
      $i++;
    }
    $i++;
    while(true){
      while(true){
        if ($i > $this->j)
          return $n;
        if ($this->cons($i))
          break;
        $i++;
      }
      $i++;
      $n++;
      while(true){
        if ($i > $this->j)
          return $n;
        if (! $this->cons($i))
          break;
        $i++;
      }
      $i++;
    }
  }


  /* vowelinstem() is true <=> 0,...j contains a vowel */
  function vowelinstem(){
    for ($i = 0; $i <= $this->j; $i++)
      if (! ($this->cons($i)))
        return true;
    return false;
  }

  /* doublec(j) is true <=> j, (j-1) contain a double consonant. */
  function doublec($j){
    if ($j < 1)
      return false;
    if ($this->buffer[$j] != $this->buffer[$j-1])
      return false;
    return $this->cons($j);
  }

  /* cvc(i) is true <=> i-2, i-1, i has the form consonant - vowel - consonant
     and also if the second c is not w, x or y. this is used when trying to
     restore an e at the end of a short word. e.g.
     cav(e), lov(e), hop(e), crim(e), but
     snow, box, tray */

  function  cvc($i){
    if ($i < 0 || !$this->cons($i) || $this->cons($i-1) || !$this->cons($i-2))
      return false;

    $ch = $this->buffer[$i];
    if ($ch == "w" || $ch == "x" || $ch == "y")
      return false;

    return true;
  }

  function ends($s){
    $l = strlen($s);
    $o = $this->k - $l + 1;
    if ($o < 0)
      return false;

    for ($i = 0; $i < $l; $i++)
      if ($this->buffer[$o + $i] != substr($s, $i, 1))
        return false;

    $this->j = $this->k - $l;
    return true;
  }

  /* setto(s) sets (j+1), ...k to the characters in the string s, readjusting k */
  function setto($s){
    $l = strlen($s);
    $o = $this->j+1;
    for ($i = 0; $i < $l; $i++)
      $this->buffer[$o+$i] = substr($s, $i, 1);
    $this->k = $this->j+$l;
  }

  function r($s) {
    if ($this->m() > 0) $this->setto($s);
  }

  /* STEP 1: gets rid of plurals and -ed or -ing. e.g. */
  function step1(){
    if ($this->buffer[$this->k] == "s"){
      if ($this->ends("sses"))
        $this->k -= 2;
      else
        if ($this->ends("ies"))
          $this->setto("i");
        else
          if ($this->buffer[$this->k-1] != "s")
            $this->k--;
    }
    if ($this->ends("eed")) {
      if ($this->m() > 0)
        $this->k--;
    } else
      if (($this->ends("ed") || $this->ends("ing")) && $this->vowelinstem()){
        $this->k = $this->j;
        if ($this->ends("at"))
          $this->setto("ate");
        else
          if ($this->ends("bl"))
            $this->setto("ble");
          else
            if ($this->ends("iz"))
              $this->setto("ize");
            else
              if ($this->doublec($this->k)){
                $this->k--;
                $ch = $this->buffer[$this->k];
                if ($ch == "l" || $ch == "s" || $ch == "z")
                  $this->k++;
              }
              else if ($this->m() == 1 && $this->cvc($this->k))
                $this->setto("e");

      }
  }

  /* STEP 2: turns terminal y to i when there is another vowel in the stem. */
  function step2() {
    if ($this->ends("y") && $this->vowelinstem())
      $this->buffer[$this->k] = "i";
  }


  /* STEP 3: maps double suffices to single ones. so -ization ( = -ize plus
     -ation) maps to -ize etc. note that the string before the suffix must give
     m() > 0 */
  function step3() {
    switch ($this->buffer[$this->k-1]){
    case "a":
      if ($this->ends("ational")) {
        $this->r("ate");
        break;
      }
      if ($this->ends("tional")) {
        $this->r("tion");
        break;
      }
      break;
    case "c":
      if ($this->ends("enci")) {
        $this->r("ence");
        break;
      }
      if ($this->ends("anci")) {
        $this->r("ance");
        break;
      }
      break;
    case "e":
      if ($this->ends("izer")) {
        $this->r("ize");
        break;
      }
      break;
    case "l":
      if ($this->ends("bli")) {
        $this->r("ble");
        break;
      }
      if ($this->ends("alli")) {
        $this->r("al");
        break;
      }
      if ($this->ends("entli")) {
        $this->r("ent"); break;
      }
      if ($this->ends("eli")) {
        $this->r("e");
        break;
      }
      if ($this->ends("ousli")) {
        $this->r("ous");
        break;
      }
      break;
    case "o":
      if ($this->ends("ization")) {
        $this->r("ize");
        break;
      }
      if ($this->ends("ation")) {
        $this->r("ate");
        break;
      }
      if ($this->ends("ator")) {
        $this->r("ate");
        break;
      }
      break;
    case "s":
      if ($this->ends("alism")) {
        $this->r("al");
        break;
      }
      if ($this->ends("iveness")) {
        $this->r("ive");
        break;
      }
      if ($this->ends("fulness")) {
        $this->r("ful");
        break;
      }
      if ($this->ends("ousness")) {
        $this->r("ous");
        break;
      }
      break;
    case "t":
      if ($this->ends("aliti")) {
        $this->r("al");
        break;
      }
      if ($this->ends("iviti")) {
        $this->r("ive");
        break;
      }
      if ($this->ends("biliti")) {
        $this->r("ble");
        break;
      }
      break;
    case "g":
      if ($this->ends("logi")) {
        $this->r("log");
        break;
      }
    }
  }

  /* STEP 4: deals with -ic-, -full, -ness etc. similar strategy to STEP 3 */
  function step4() {
    switch ($this->buffer[$this->k]){
    case "e":
      if ($this->ends("icate")) {
        $this->r("ic");
        break;
      }
      if ($this->ends("ative")) {
        $this->r("");
        break;
      }
      if ($this->ends("alize")) {
        $this->r("al");
        break;
      }
      break;
    case "i":
      if ($this->ends("iciti")) {
        $this->r("ic");
        break;
      }
      break;
    case "l":
      if ($this->ends("ical")) {
        $this->r("ic");
        break;
      }
      if ($this->ends("ful")) {
        $this->r("");
        break;
      }
      break;
    case "s":
      if ($this->ends("ness")) {
        $this->r("");
        break;
      }
      break;
    }
  }


  /* STEP 5: takes off -ant, -ence etc., in context <c>vcvc<v> */
  function step5(){
    switch ($this->buffer[$this->k-1]){
    case "a":
      if ($this->ends("al"))
        break;
      return;
    case "c":
      if ($this->ends("ance"))
        break;
      if ($this->ends("ence"))
        break;
      return;
    case "e":
      if ($this->ends("er"))
        break;
      return;
    case "i":
      if ($this->ends("ic"))
        break;
      return;
    case "l":
      if ($this->ends("able"))
        break;
      if ($this->ends("ible"))
        break;
      return;
    case "n":
      if ($this->ends("ant"))
        break;
      if ($this->ends("ement"))
        break;
      if ($this->ends("ment"))
        break;
      /* element etc. not stripped before the m */
      if ($this->ends("ent"))
        break;
      return;
    case "o":
      if ($this->ends("ion") &&
          $this->buffer[$this->j] == "s" ||
          $this->buffer[$this->j] == "t")
        break;
      if ($this->ends("ou"))
        break;
      return;
      /* takes care of -ous */
    case "s":
      if ($this->ends("ism"))
        break;
      return;
    case "t":
      if ($this->ends("ate"))
        break;
      if ($this->ends("iti"))
        break;
      return;
    case "u":
      if ($this->ends("ous"))
        break;
      return;
    case "v":
      if ($this->ends("ive"))
        break;
      return;
    case "z":
      if ($this->ends("ize"))
        break;
      return;
    default:
      return;
    }
    if ($this->m() > 1) $this->k = $this->j;
  }

  /* STEP 6: removes a final -e if m() > 1 */
  function step6(){
    $this->j = $this->k;
    if ($this->buffer[$this->k] == "e"){
      $a = $this->m();
      if ($a > 1 || $a == 1 && !$this->cvc($this->k-1))
        $this->k--;
    }
    if ($this->buffer[$this->k] == "l" && $this->doublec($this->k) && $this->m() > 1)
      $this->k--;
  }

  function stemstr($string)  {
    $string = strtolower($string);
    for($i = 0; $i < strlen($string); $i++)
      $this->buffer[$this->stemIdx++] = substr($string, $i, 1);

    $this->k = $this->stemIdx - 1;
    if ($this->k > 1){
      $this->step1();
      if($this->k > 2)
        $this->step2();
      if($this->k > 2)
        $this->step3();
      if($this->k > 2)
        $this->step4();
      if($this->k > 2)
        $this->step5();
      if($this->k > 2)
        $this->step6();
    }
    $this->stemIdx = $this->k + 1;

    $stem = implode("", $this->buffer);
    $stem = (substr($stem, 0, $this->stemIdx));
    $this->stemIdx = 0;
    $this->buffer = array();
    return $stem;
  }
}

?>
